//********************************************************************
UNIT Shape3DUnit;
{******************************************************************}
(*
  Released as part of the DGLSDK
  Visit us @ http://www.delphigl.com
  Maintained by Jonas Scholz (i0n0s)
  jonasscholz@web.de
-------------------------------------------------------------------
  License
  -------
  Copyright (C) 2009 www.delphigl.com

  This license applies to everything on www.delphigl.com, except where otherwise noted.

  This software is provided 'as-is', without any express or implied warranty.  In no event will the authors
  be held liable for any damages arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose, including commercial applications,
  and to alter it and redistribute it freely, subject to the following restrictions:

  1. The origin of this software must not be misrepresented; you must not claim that you wrote the original
  software. If you use this software in a product, an acknowledgment in the product documentation would be
  appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.

  Description
  -----------
  Models for the easySDL-framework, see german description below

  Requirements
  ------------
  - SDL 1.2 (http://www.libsdl.org/download-1.2.php or your distribution)
  - Freepascal 2.0.0 or above (http://www.freepascal.org/download.html)
  - JEDI-SDL 1.0 (http://www.delphi-jedi.org/Jedi:TEAM_SDL_HOME)
  - DGLSDK for Linux (http://www.delphigl.com/)


  History
  -------
  Created January 2009 (GT)

  Contact
  -------
  We created this source code in our spare time. If you find a bug or just
  wanna say hello, feel free to write to the contributors:

  GT
    Gertraude Trawnicek (Traude)
    send me a PM at www.delphigl.com

*******************************************************************************)
// Beschreibung:
// -------------
// TShape3D kann eine Kugel oder einen Torus erzeugen und anzeigen.
//
// Der CONSTRUCTOR erwartet einen Shape-Typ (stSphere oder stTorus)
// und die Anzahl der Zeilen und Spalten, die das Gebilde haben soll.
//
// Die RENDER-Prozedur erwartet einen Parameter, ob sie mit Face-
// Normalen (Smooth=FALSE) oder mit Vertex-Normalen (Smooth=TRUE)
// zeichnen soll.
//
// Texturkoordinaten (S,T) sind dabei.
//********************************************************************
INTERFACE

USES Math, dglOpenGl;

TYPE
  TVertexIndex = Integer;
  TVector3D = Record X,Y,Z: Single; End;
  TVertexField = Array Of TVector3D;

  TTexCoord2D = Record S,T: Single; End;
  TTexCoordField = Array Of TTexCoord2D;

  TShape3DType = (stSphere, stTorus);

  TShape3D = Class(TObject)
  Private
     fShape3DType: TShape3DType;
     fColCount: Cardinal;
     fRowCount: Cardinal;
     fAnglePerCol: Word;
     fAnglePerRow: Word;

     fVertex: TVertexfield;
     fVxNormal: TVertexfield;
     fFaceNormal: TVertexField;
     fTexCoord: TTexCoordField;

     Procedure CreateVerticesAndTexCoords;
     Procedure CreateFaceNormals;
     Procedure CreateVertexNormals;

     Procedure Initialize;
     Function CreateVertex(ACol,ARow: Cardinal): TVector3D;
     Function GetLRFaceIndex(ACol,ARow: TVertexIndex): TVertexIndex;
     Function GetVertexIndex(ACol,ARow: TVertexIndex): TVertexIndex;
     Function GetVertexCount: TVertexIndex;
     Function GetFaceCount: TVertexIndex;
     Function CalcVertexNormal(ACol,ARow: Cardinal): TVector3D;
     Function CorrectInput(AInput, ADividend: Word): Word;
     Function ValidIndex(ACol,ARow: Cardinal): Boolean;
  Public
     Constructor Create(AShape3DType: TShape3DType;
                AColCount,ARowCount: Cardinal); Reintroduce;
     Destructor Destroy; Override;
     Procedure Render(Smooth: Boolean);
  End;

//********************************************************************
IMPLEMENTATION
//********************************************************************
CONST
   PiDiv180 = PI/180;
   Tolerance = 0.001;
   ErrorVector: TVector3D = (X:NaN; Y:NaN; Z:NaN);

Function VectorCrossProduct(Vector1,Vector2: TVector3D): TVector3D; Forward;
Function VectorDeduct(FromPoint,ToPoint: TVector3D): TVector3D; Forward;
Function VectorIsNaN(AVector: TVector3D): Boolean; Forward;
Function VectorUnify(AVector: TVector3D): TVector3D; Forward;
//********************************************************************
Constructor TShape3D.Create(AShape3DType: TShape3DType;
          AColCount,ARowCount: Cardinal);
Begin
   Inherited Create;

   fShape3DType:= AShape3DType;
   fColCount:= AColCount;
   fRowCount:= ARowCount;

   Initialize;

   SetLength(fVertex,GetVertexCount);
   SetLength(fTexCoord,GetVertexCount);
   CreateVerticesAndTexCoords;

   SetLength(fFaceNormal,GetFaceCount);
   CreateFaceNormals;

   // Achtung: die Vertexnormalen knnen erst NACH den Facenormalen
   // berechnet werden
   SetLength(fVxNormal,GetVertexCount);
   CreateVertexNormals;
End;
//********************************************************************
Destructor TShape3D.Destroy;
Begin
   Finalize(fFaceNormal);
   Finalize(fVxNormal);
   Finalize(fTexCoord);
   Finalize(fVertex);

   Inherited Destroy;
End;
//********************************************************************
Procedure TShape3D.Initialize;
Begin
  Case fShape3DType Of

     stSphere: Begin
           fColCount:= CorrectInput(fColCount,360);
           fRowCount:= CorrectInput(fRowCount,180);

           fAnglePerCol:= 360 Div fColCount;
           fAnglePerRow:= 180 Div fRowCount;
        End;

     stTorus: Begin
           fColCount:= CorrectInput(fColCount,360);
           fRowCount:= CorrectInput(fRowCount,360);

           fAnglePerCol:= 360 Div fColCount;
           fAnglePerRow:= 360 Div fRowCount;
        End;
  End;
End;
//********************************************************************
Procedure TShape3D.CreateVerticesAndTexCoords;
Var Col,Row,Index: TVertexIndex;
Begin
   // Es muss eine Spalte und eine Zeile mehr als angegegeben
   // erzeugt werden
   For Row:= 0 To fRowCount Do
      For Col:= 0 To fColCount Do Begin

         Index:= GetVertexIndex(Col,Row);
         fVertex[Index]:= CreateVertex(Col,Row);
         
         With fTexCoord[Index] Do Begin
            S:= 1-Col/fColCount;
            T:= 1-Row/fRowCount;
         End;
      End;
End;
//********************************************************************
Procedure TShape3D.CreateFaceNormals;
Var FaceIndex,Col,Row: TVertexIndex;
Begin
   FaceIndex:= 0;

   For Row:= 0 To fRowCount-1 Do
      For Col:= 0 To fColCount-1 Do Begin

         // Vektoren: A=>B, A=>C
         fFaceNormal[(FaceIndex)]:= VectorCrossProduct
            (VectorDeduct(fVertex[GetVertexIndex(Col  ,Row  )],
                          fVertex[GetVertexIndex(Col+1,Row  )]),
             VectorDeduct(fVertex[GetVertexIndex(Col  ,Row  )],
                          fVertex[GetVertexIndex(Col+1,Row+1)]));

         fFaceNormal[(FaceIndex)]:=
            VectorUnify(fFaceNormal[(FaceIndex)]);

         If VectorIsNan(fFaceNormal[(FaceIndex)]) Then Begin
         // Dies betrifft nur die erste Facereihe der Kugel
         // bei der die Punkte A und B identisch sind
         // Vektoren: A=>C, A=>D
         fFaceNormal[(FaceIndex)]:= VectorCrossProduct
            (VectorDeduct(fVertex[GetVertexIndex(Col  ,Row  )],
                          fVertex[GetVertexIndex(Col+1,Row+1)]),
             VectorDeduct(fVertex[GetVertexIndex(Col  ,Row  )],
                          fVertex[GetVertexIndex(Col  ,Row+1)]));

            fFaceNormal[(FaceIndex)]:=
               VectorUnify(fFaceNormal[(FaceIndex)]);
         End;

         Inc(FaceIndex);
      End;
End;
//********************************************************************
Procedure TShape3D.CreateVertexNormals;
Var Col,Row: TVertexIndex;
Begin
   For Row:= 0 To fRowCount Do
      For Col:= 0 To fColCount Do
         fVxNormal[GetVertexIndex(Col,Row)]:= CalcVertexNormal(Col,Row);
End;
//********************************************************************
Function TShape3D.CreateVertex(ACol,ARow: Cardinal): TVector3D;

//--------------------------------------------------------------------
Function SpherePoint: TVector3D;
Begin
   With Result Do Begin
      X:= Sin(fAnglePerRow*ARow*PiDiv180) * Cos(fAnglePerCol*ACol*PiDiv180);
      Y:= Cos(fAnglePerRow*ARow*PiDiv180);
      Z:= Sin(fAnglePerRow*ARow*PiDiv180) * Sin(fAnglePerCol*ACol*PiDiv180);
   End;
End;
//--------------------------------------------------------------------
Function TorusPoint: TVector3D;
Const Radius = 0.3;
Begin
   With Result Do Begin
      X:= Cos(fAnglePerCol*ACol*PiDiv180)*(1+Radius*Cos(fAnglePerRow*ARow*PiDiv180));
      Y:= Radius*Sin(fAnglePerRow*ARow*PiDiv180);
      Z:= -Sin(fAnglePerCol*ACol*PiDiv180)*(1+Radius*Cos(fAnglePerRow*ARow*PiDiv180));
   End;
End;
//--------------------------------------------------------------------
Begin
  Case fShape3DType Of
     stSphere: Result:= SpherePoint;
      stTorus: Result:= TorusPoint;
  End;
End;
//********************************************************************
Procedure TShape3D.Render(Smooth: Boolean);
Var FaceIndex,Col,Row: TVertexIndex;
Begin
   FaceIndex:= 0;

   glBegin(GL_QUADS);
      For Row:= 0 To fRowCount-1 Do
         For Col:= 0 To fColCount-1 Do Begin

            If Not Smooth
               Then With fFaceNormal[FaceIndex] Do glNormal3f(X,Y,Z);

            If Smooth
               Then With fVxNormal[GetVertexIndex(Col  ,Row  )] Do glNormal3f(X,Y,Z);
            With fTexCoord[GetVertexIndex(Col  ,Row  )] Do glTexCoord2f(S,T);
            With fVertex[GetVertexIndex(Col  ,Row  )] Do glVertex3f(X,Y,Z);

            If Smooth
               Then With fVxNormal[GetVertexIndex(Col+1,Row  )] Do glNormal3f(X,Y,Z);
            With fTexCoord[GetVertexIndex(Col+1,Row  )] Do glTexCoord2f(S,T);
            With fVertex[GetVertexIndex(Col+1,Row  )] Do glVertex3f(X,Y,Z);

            If Smooth
               Then With fVxNormal[GetVertexIndex(Col+1,Row+1)] Do glNormal3f(X,Y,Z);
            With fTexCoord[GetVertexIndex(Col+1,Row+1)] Do glTexCoord2f(S,T);
            With fVertex[GetVertexIndex(Col+1,Row+1)] Do glVertex3f(X,Y,Z);

            If Smooth
               Then With fVxNormal[GetVertexIndex(Col  ,Row+1)] Do glNormal3f(X,Y,Z);
            With fTexCoord[GetVertexIndex(Col  ,Row+1)] Do glTexCoord2f(S,T);
            With fVertex[GetVertexIndex(Col  ,Row+1)] Do glVertex3f(X,Y,Z);

            Inc(FaceIndex);
         End;
   glEnd();
End;
//********************************************************************
// Gibt den Index des Lower Right Faces zurck (bezogen auf das
// Vertexfeld)
Function TShape3D.GetLRFaceIndex(ACol,ARow: TVertexIndex): TVertexIndex;
Begin
   Result:= -1;
   If ValidIndex(ACol,ARow)
      Then Result:= fColCount*ARow+ACol;
End;
//********************************************************************
Function TShape3D.GetVertexIndex(ACol,ARow: TVertexIndex): TVertexIndex;
Begin
   // das Vertexfeld ist eine Spalte breiter und eine Zeile hher
   // als angegeben
   Result:= -1;
   If ValidIndex(ACol,ARow)
      Then Result:= (fColCount+1)*ARow+ACol;
End;
//********************************************************************
Function TShape3D.GetVertexCount: TVertexIndex;
Begin
   Result:= (fColCount+1)*(fRowCount+1);
End;
//********************************************************************
Function TShape3D.GetFaceCount: TVertexIndex;
Begin
   Result:= fColCount*fRowCount;
End;
//********************************************************************
// Berechnet die Vertex-Normalen auf Basis der 4 angrenzenden Quads
Function TShape3D.CalcVertexNormal(ACol,ARow: Cardinal): TVector3D;
Var

   UpperLeft,UpperRight,LowerLeft,LowerRight: TVector3D;
   LRFaceIndex: TVertexIndex;
Begin
   Result:= ErrorVector;

   If ValidIndex(ACol,ARow) Then Begin

      LRFaceIndex:= GetLRFaceIndex(ACol,ARow);
      LowerRight:= fFaceNormal[LRFaceIndex];

      If (ARow > 0)
         Then UpperRight:= fFaceNormal[LRFaceIndex-fColCount]
         Else UpperRight:= LowerRight;

      If (ACol > 0) And (ARow > 0)
         Then UpperLeft:= fFaceNormal[LRFaceIndex-fColCount-1]
         Else UpperLeft:= LowerRight;

      If (ACol > 0)
         Then LowerLeft:= fFaceNormal[LRFaceIndex-1]
         Else LowerLeft:= LowerRight;

      With Result Do Begin
         X:= (UpperLeft.X+UpperRight.X+LowerLeft.X+LowerRight.X)/4;
         Y:= (UpperLeft.Y+UpperRight.Y+LowerLeft.Y+LowerRight.Y)/4;
         Z:= (UpperLeft.Z+UpperRight.Z+LowerLeft.Z+LowerRight.Z)/4;
      End;
   End;
End;
//********************************************************************
Function TShape3D.ValidIndex(ACol,ARow: Cardinal): Boolean;
Begin
   If (ACol <= fColCount) And (ARow <= fRowCount)
      Then Result:= TRUE
      Else Result:= FALSE;
End;
//********************************************************************
Function TShape3D.CorrectInput(AInput, ADividend: Word): Word;
Var Temp: Single;
Begin
   // Prfen, ob die Division Dividend/Input einen Rest ergibt
   If AInput <= 0
      Then AInput:= 8;
   If AInput > 360
      Then AInput:= 360;

   Dec(AInput);
   Repeat
      Inc(AInput);
      Temp:= ADividend/AInput;
   Until (Temp-Trunc(Temp) = 0) And (Not Odd(AInput));

   Result:= AInput;
End;
//********************************************************************
//********************************************************************
//********************************************************************
//********************************************************************
//********************************************************************
// Vector-Utilities
//********************************************************************
Function VectorCrossProduct(Vector1,Vector2: TVector3D): TVector3D;
Begin
   Result.X:= Vector1.Y*Vector2.Z - Vector1.Z*Vector2.Y;
   Result.Y:= Vector1.Z*Vector2.X - Vector1.X*Vector2.Z;
   Result.Z:= Vector1.X*Vector2.Y - Vector1.Y*Vector2.X;
End;
//********************************************************************
Function VectorDeduct(FromPoint,ToPoint: TVector3D): TVector3D;
Begin
   Result.X:= ToPoint.X - FromPoint.X;
   Result.Y:= ToPoint.Y - FromPoint.Y;
   Result.Z:= ToPoint.Z - FromPoint.Z;
End;
//********************************************************************
Function VectorIsNaN(AVector: TVector3D): Boolean;
Begin
   Result:= FALSE;

   With AVector Do
      If IsNan(X) Or IsNan(Y) Or IsNan(Z)
         Then Result:= TRUE;
End;
//********************************************************************
Function VectorUnify(AVector: TVector3D): TVector3D;
Var LVectorValue: Single;
Begin
   With AVector Do LVectorValue:= Sqrt(X*X+Y*Y+Z*Z);
   If LVectorValue > Tolerance
      Then With Result Do Begin
         X:= AVector.X / LVectorValue;
         Y:= AVector.Y / LVectorValue;
         Z:= AVector.Z / LVectorValue;
      End
      Else With Result Do Begin
         X:= NaN;
         Y:= NaN;
         Z:= NaN;
      End;
End;
//********************************************************************
//********************************************************************
End.
